#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2018
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
from asyncio.log import logger
import functools
import json
import os
import pytest
import unittest

import numpy as np

from dlg.dask_emulation import delayed as dlg_delayed
from dlg.dask_emulation import compute as dlg_compute
from dlg.common import tool
from dlg.dask_emulation import compute as dlg_compute
from dlg.dask_emulation import delayed as dlg_delayed
from dlg.utils import terminate_or_kill
from six.moves import reduce  # @UnresolvedImport

try:
    from dask import delayed as dask_delayed
    from dask import compute as dask_compute
except ImportError:
    dask_delayed = None


def add(x, y):
    return x + y


def add_list(numbers):
    return functools.reduce(add, numbers)


def subtract(x, y):
    return x - y


def subtract_list(numbers):
    return functools.reduce(subtract, numbers)


def multiply(x, y):
    return x * y


def divide(x, y):
    return x / y


def partition(x):
    return x / 2, x - x / 2


def sum_with_args(a, a1=0, a2=0, a3=0):
    """Returns a + sum((a1, a2, a3))"""
    return a + sum([a1, a2, a3])


def sum_with_kwargs(a, **kwargs):
    """Returns a + kwargs['b'], or only a if no 'b' is found in kwargs"""
    b = kwargs.pop("b", 0)
    return a + b


def sum_with_args_and_kwarg(a, *args, **kwargs):
    """Returns a + sum(args) + b, or only a +sum(args) if no 'b' is found in kwargs"""
    b = kwargs.pop("b", 0)
    return a + sum(args) + b


class MyType(object):
    """A type that is serializable but not convertible to JSON"""

    def __init__(self, x):
        self.x = x
        self.array = np.zeros(1)


try:
    json.dumps(MyType(1))
    assert False, "Should fail to serialize to JSON"
except:
    pass


def sum_with_user_defined_default(a, b=MyType(10)):
    return a + b.x


@pytest.fixture(scope="function", autouse=True)
def change_test_dir(request, monkeypatch):
    print(f">>>>> path: {request.fspath.dirname}")
    monkeypatch.chdir(request.fspath.dirname)


class _TestDelayed(object):
    """Test definitions run under non-delayed, dlg_delayed and possibly dask_delayed contexts"""

    def test_simple(self):
        """This creates the following graph

        1 --|
            |-- add --> the_sum --------|
        2 --|                           |                                       |--> part1 --|
                                        |--> divide --> division -->partition --|            |--> add -> final
        4 --|                           |                                       |--> part2 --|
            |-- substract --> the_sub --|
        3 --|
        """
        delayed = self.delayed
        compute = self.compute

        the_sum = delayed(add)(1.0, 2.0)
        the_sub = delayed(subtract)(4.0, 3.0)
        division = delayed(divide)(the_sum, the_sub)
        parts = delayed(partition, nout=2)(division)
        logger.debug(f"partitions: {type(parts)}")
        result = compute(delayed(add)(*parts))
        self.assertEqual(3.0, result)

    def test_args_as_lists(self):
        """Like test_simple, but some arguments are passed down as lists"""
        delayed = self.delayed
        compute = self.compute

        one, two, three, four = (
            delayed(1.0),
            delayed(2.0),
            delayed(3.0),
            delayed(4.0),
        )
        the_sum = delayed(add_list)([one, two])
        the_sub = delayed(subtract_list)([four, three])
        division = delayed(divide)(the_sum, the_sub)
        parts = delayed(partition, nout=2)(division)
        result = compute(delayed(add)(*parts))
        self.assertEqual(3.0, result)

    def test_compute_with_lists(self):
        """Make sure we can call compute() directly on the list objects"""
        delayed = self.delayed
        compute = self.compute

        one, two, three, four = (
            delayed(1.0),
            delayed(2.0),
            delayed(3.0),
            delayed(4.0),
        )
        doubles = [delayed(lambda i: i * 2)(x) for x in (one, two, three, four)]
        result = compute(doubles)
        self.assertEqual([2.0, 4.0, 6.0, 8.0], result)

    def test_none_arg(self):
        """Test that calling delayed(f)(None) works"""
        delayed = self.delayed
        compute = self.compute
        self.assertEqual(compute(delayed(lambda _: None)(None)), None)

    def test_with_args(self):
        """Tests that delayed() works correctly with positional args"""
        delayed = self.delayed
        compute = self.compute
        logger.info(f"Running compute(delayed(sum_with_args)(1)), 1")
        self.assertEqual(compute(delayed(sum_with_args)(1)), 1)
        logger.info(f"Running compute(delayed(sum_with_args)(1, 20)), 21")
        self.assertEqual(compute(delayed(sum_with_args)(1, 20)), 21)
        logger.info(f"Running compute(delayed(sum_with_args)(1, 20, 30)), 51")
        self.assertEqual(compute(delayed(sum_with_args)(1, 20, 30)), 51)

    def test_with_kwargs(self):
        """Tests that delayed() works correctly with args and kwargs"""
        delayed = self.delayed
        compute = self.compute

        self.assertEqual(compute(delayed(sum_with_kwargs)(1)), 1)

    def test_with_args_and_kwargs(self):
        """Tests that delayed() works correctly with kwargs"""
        delayed = self.delayed
        compute = self.compute

        self.assertEqual(compute(delayed(sum_with_args_and_kwarg)(1)), 1)

    def test_with_user_defined_default(self):
        """Tests that delayed() works with default values that are not json-dumpable"""
        delayed = self.delayed
        compute = self.compute

        self.assertEqual(compute(delayed(sum_with_user_defined_default)(1)), 11)
        self.assertEqual(
            compute(delayed(sum_with_user_defined_default)(1, MyType(20))), 21
        )

    def test_with_noniterable_nout_1(self):
        """Tests that using nout=1 works as expected with non-iterable objects"""
        # Simple call
        delayed = self.delayed
        compute = self.compute

        self.assertEqual(compute(delayed(add, nout=1)(1, 2)), 3)

        # Compute a delayed that uses a delayed with nout=1
        addition = delayed(add, nout=1)(1, 2)
        self.assertEqual(compute(delayed(add)(addition, 3)), 6)

        # Like above, but the first delayed also uses nout=1
        self.assertEqual(compute(delayed(add, nout=1)(addition, 3)), 6)

    def test_with_iterable_nout_1(self):
        """Tests that using nout=1 works as expected with iterable objects"""
        # Like last test above, but result is actually iterable
        delayed = self.delayed
        compute = self.compute

        listify = lambda _x: [_x]
        for x in delayed(listify, nout=1)(1):
            results = compute(delayed(add)(x, 2))
        self.assertEqual(results, 3)


class TestNoDelayed(unittest.TestCase, _TestDelayed):
    """Non-delayed tests"""

    def delayed(self, f, *_, **__):
        return f

    def compute(self, val):
        return val


class TestDlgDelayed(_TestDelayed, unittest.TestCase):
    """dlg-base tests, they start/stop the node manager and use dlg_delayed"""

    def delayed(self, f, *args, **kwargs):
        return dlg_delayed(f, *args, **kwargs)

    def setUp(self):
        env = os.environ.copy()
        env["PYTHONPATH"] = f"{env.get('PYTHONPATH', '')}:{os.path.abspath('.')}/.."
        print(f">>>> env: {env['PYTHONPATH']}")
        unittest.TestCase.setUp(self)
        self.dmProcess = tool.start_process("nm", ["-vvv"], env=env)

    def compute(self, val):
        return dlg_compute(val)

    def tearDown(self):
        terminate_or_kill(self.dmProcess, 5)
        unittest.TestCase.tearDown(self)


@unittest.skipIf(dask_delayed is None, "dask is not available")
class TestDaskDelayed(_TestDelayed, unittest.TestCase):
    """dask-base tests, they use dask_delayed"""

    def delayed(self, f, *args, **kwargs):
        return dask_delayed(f, *args, **kwargs)

    def compute(self, val):
        logger.info(f"Running compute...")
        return dask_compute(val)[0]



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import subprocess
import unittest

from dlg import common
from dlg.common import tool
from dlg.testutils import ManagerStarter


class TestTool(ManagerStarter, unittest.TestCase):
    def test_cmdhelp(self):
        """Checks that all dlg commands have a help"""
        tool._load_commands()
        for cmd in tool.commands:
            p = tool.start_process(
                cmd, ["-h"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            out, err = p.communicate()
            common.wait_or_kill(p, timeout=10)
            self.assertEqual(
                0,
                p.returncode,
                "cmd: %s, out: %s" % (cmd + " -h", common.b2s(out + err)),
            )



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import logging
import threading

from six.moves import queue as Queue  # @UnresolvedImport

from dlg.apps.app_base import BarrierAppDROP


logger = logging.getLogger(__name__)


class SourceFlux(BarrierAppDROP):
    def initialize(self, **kwargs):

        super(SourceFlux, self).initialize(**kwargs)

        self.casapy_path = self._popArg(kwargs, "casapy_path", None)
        self.timeout = self._popArg(kwargs, "timeout", 180)

    def run(self):
        inp = self.inputs[0]
        out = self.outputs[0]

        if logger.isEnabledFor(logging.INFO):
            logger.info("Calculating source flux on %s.image", inp.path)

        import drivecasa

        casa = drivecasa.Casapy(casa_dir=self.casapy_path, timeout=self.timeout)
        casa.run_script(['ia.open("' "%s" '")' % (inp.path + ".image")])
        casa.run_script(
            ['flux = ia.pixelvalue([128,128,0,179])["' "value" '"]["' "value" '"]']
        )
        casaout, _ = casa.run_script(["print flux"])
        flux = float(casaout[0])
        if flux > 9e-4:
            if logger.isEnabledFor(logging.INFO):
                logger.info("Valid flux found: %f", flux)
        out.write(str(flux))


class Clean(BarrierAppDROP):
    def initialize(self, **kwargs):

        super(Clean, self).initialize(**kwargs)

        self.timeout = self._popArg(kwargs, "timeout", 3600)
        self.casapy_path = self._popArg(kwargs, "casapy_path", None)

        self.clean_args = {
            "field": str(self._popArg(kwargs, "field", None)),
            "mode": str(self._popArg(kwargs, "mode", None)),
            "restfreq": str(self._popArg(kwargs, "restfreq", None)),
            "nchan": self._popArg(kwargs, "nchan", None),
            "start": str(self._popArg(kwargs, "start", None)),
            "width": str(self._popArg(kwargs, "width", None)),
            "interpolation": str(self._popArg(kwargs, "interpolation", None)),
            "gain": self._popArg(kwargs, "gain", None),
            "imsize": self._popArg(kwargs, "imsize", None),
            "cell": [str(x) for x in self._popArg(kwargs, "cell", [])],
            "phasecenter": str(self._popArg(kwargs, "phasecenter", None)),
            "weighting": str(self._popArg(kwargs, "weighting", None)),
            "usescratch": False,
        }

    def invoke_clean(self, q, vis, outcube):

        import drivecasa

        try:
            script = []
            casa = drivecasa.Casapy(casa_dir=self.casapy_path, timeout=self.timeout)
            drivecasa.commands.clean(
                script,
                vis_paths=vis,
                out_path=outcube,
                niter=0,
                threshold_in_jy=0,
                other_clean_args=self.clean_args,
                overwrite=True,
            )
            casa.run_script(script)
            q.put(0)

        except Exception:
            q.put(-1)
            raise

    def run(self):

        inp = self.inputs
        out = self.outputs[0]

        vis = [i.path for i in inp]

        if logger.isEnabledFor(logging.INFO):
            logger.info("Cleaning %r", vis)

        q = Queue.Queue()
        t = threading.Thread(target=self.invoke_clean, args=(q, vis, out.path))
        t.start()
        t.join()

        result = q.get()
        if result != 0:
            raise Exception("Error cleaning")


class Split(BarrierAppDROP):
    def initialize(self, **kwargs):

        super(Split, self).initialize(**kwargs)

        self.timeout = self._popArg(kwargs, "timeout", 3600)
        self.casapy_path = self._popArg(kwargs, "casapy_path", False)

        self.transform_args = {
            "regridms": self._popArg(kwargs, "regridms", None),
            "restfreq": str(self._popArg(kwargs, "restfreq", None)),
            "mode": str(self._popArg(kwargs, "mode", None)),
            "nchan": self._popArg(kwargs, "nchan", None),
            "outframe": str(self._popArg(kwargs, "outframe", None)),
            "interpolation": str(self._popArg(kwargs, "interpolation", None)),
            "veltype": "radio",
            "start": str(self._popArg(kwargs, "start", None)),
            "width": str(self._popArg(kwargs, "width", None)),
            "spw": "",
            "combinespws": True,
            "nspw": 1,
            "createmms": False,
            "datacolumn": "data",
        }

    def invoke_split(self, q, infile, outdir):

        import drivecasa

        try:
            script = []
            casa = drivecasa.Casapy(casa_dir=self.casapy_path, timeout=self.timeout)
            drivecasa.commands.mstransform(
                script, infile, outdir, self.transform_args, overwrite=True
            )
            casa.run_script(script)
            q.put(0)

        except Exception:
            q.put(-1)
            raise

    def run(self):
        inp = self.inputs[0]
        out = self.outputs[0]

        if logger.isEnabledFor(logging.INFO):
            logger.info("Splitting %s", inp.path)

        q = Queue.Queue()
        t = threading.Thread(target=self.invoke_split, args=(q, inp.path, out.path))
        t.start()
        t.join()

        result = q.get()
        if result != 0:
            raise Exception("Error splitting")



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import json
import os

from dlg.common import dropdict


# Directories and paths
CASAPY = "/home/ec2-user/casa-release-4.4.0-el6/"
INPUTS_DIR = "/home/ec2-user/data/input"
OUTPUTS_DIR = "/home/ec2-user/data/output"
KEY_PATH = "/home/ec2-user/.ssh/aws-sdp-sydney.pem"

# Internal AWS IP addresses. ch05 does the Clean-ing
ch01 = "172.31.4.12"
ch02 = "172.31.9.163"
ch03 = "172.31.11.184"
ch04 = "172.31.11.87"
ch05 = "172.31.0.36"

# Per-visibility parameters and resulting cube name
VIS = [
    (
        "20131025_951_4_FINAL_PRODUCTS/20131025_951_4_calibrated_deepfield.ms",
        "20131025_951_4",
        ch01,
    ),
    (
        "20140105_946_6_FINAL_PRODUCTS/20140105_946_6_calibrated_deepfield.ms",
        "20140105_946_6",
        ch02,
    ),
    (
        "20131031_951_4_FINAL_PRODUCTS/20131031_951_4_calibrated_deepfield.ms",
        "20131031_951_4",
        ch03,
    ),
    (
        "20131121_946_6_FINAL_PRODUCTS/20131121_946_6_calibrated_deepfield.ms",
        "20131121_946_6",
        ch04,
    ),
]
CUBE_NAME = "cube1408~1412"


def fileSpec(uid, **kwargs):
    dropSpec = dropdict(
        {
            "oid": str(uid),
            "categoryType": "Data",
            "dropclass": "dlg.data.drops.file.FileDROP",
        }
    )
    dropSpec.update(kwargs)
    return dropSpec


def directorySpec(uid, **kwargs):
    dropSpec = dropdict(
        {
            "oid": str(uid),
            "categoryType": "container",
            "container": "dlg.drop.DirectoryContainer",
        }
    )
    dropSpec.update(kwargs)
    return dropSpec


def fluxSpec(uid, **kwargs):
    dropSpec = dropdict(
        {
            "oid": str(uid),
            "categoryType": "Application",
            "dropclass": "test.integrate.chiles.chilesdo.SourceFlux",
        }
    )
    dropSpec.update(kwargs)
    return dropSpec


def splitSpec(uid, **kwargs):
    dropSpec = dropdict(
        {
            "oid": str(uid),
            "categoryType": "Application",
            "dropclass": "test.integrate.chiles.chilesdo.Split",
        }
    )
    dropSpec.update(kwargs)
    return dropSpec


def scpSpec(uid, **kwargs):
    dropSpec = dropdict(
        {
            "oid": str(uid),
            "categoryType": "Application",
            "dropclass": "dlg.apps.scp.ScpApp",
        }
    )
    dropSpec.update(kwargs)
    return dropSpec


def cleanSpec(uid, **kwargs):
    dropSpec = dropdict(
        {
            "oid": str(uid),
            "categoryType": "Application",
            "dropclass": "test.integrate.chiles.chilesdo.Clean",
        }
    )
    dropSpec.update(kwargs)
    return dropSpec


if __name__ == "__main__":
    droplist = []

    flux_out = fileSpec("Flux", node=ch05, dirname=os.path.join(OUTPUTS_DIR))
    droplist.append(flux_out)
    flux = fluxSpec("FluxExtractor", casapy_path=CASAPY, node=ch05)
    droplist.append(flux)

    cl = cleanSpec(
        "Cleaning",
        field="deepfield",
        mode="frequency",
        restfreq="1420.405752MHz",
        nchan=-1,
        start="",
        width="",
        interpolation="nearest",
        gain=0.1,
        imsize=[256],
        cell=["1.0arcsec"],
        phasecenter="10h01m53.9,+02d24m52s",
        weighting="natural",
        casapy_path=CASAPY,
        node=ch05,
    )

    droplist.append(cl)

    image_out = directorySpec(
        "CleanedImage",
        dirname=os.path.join(OUTPUTS_DIR, CUBE_NAME),
        check_exists=False,
        node=ch05,
    )
    droplist.append(image_out)
    cl.addOutput(image_out)
    flux.addInput(image_out)
    flux.addOutput(flux_out)

    for i, v in enumerate(VIS):
        visDir = os.path.join(INPUTS_DIR, v[0])
        splitOutDir = os.path.join(OUTPUTS_DIR, v[1])
        splitCopyDir = os.path.join(INPUTS_DIR, v[1])
        node = v[2]

        # vis -> SPLIT -> out -> scp -> out
        vis_in = directorySpec("vis%d" % (i), dirname=visDir, node=node)
        sp = splitSpec(
            "Splitting_%d" % (i),
            regridms=True,
            restfreq="1420.405752MHz",
            mode="frequency",
            nchan=256,
            outframe="lsrk",
            interpolation="linear",
            start="1408MHz",
            width="1412kHz",
            casapy_path=CASAPY,
            node=node,
        )
        split_out = directorySpec(
            "SplitOutput_%d" % (i),
            dirname=splitOutDir,
            check_exists=False,
            node=node,
        )
        scp = scpSpec("Scp_%d" % (i), node=node, pkeyPath=KEY_PATH, timeout=3600)
        scpOut = directorySpec(
            "SplitOutput_%d_Copy" % (i),
            dirname=splitCopyDir,
            check_exists=False,
            node=ch05,
        )

        # Establish relationships
        sp.addInput(vis_in)
        sp.addOutput(split_out)
        scp.addInput(split_out)
        scp.addOutput(scpOut)
        cl.addInput(scpOut)

        # Add to ifnal list of DROPs
        droplist.append(vis_in)
        droplist.append(sp)
        droplist.append(split_out)
        droplist.append(scp)
        droplist.append(scpOut)

    print(json.dumps(droplist, indent=2))
